@@ -177,6 +177,69 @@ static inline void ino_index_unlock(struct mapent_cache *mc)
 	return;
 }
 
+/* Save the cache entry mapent field onto a stack and set a new mapent */
+int cache_push_mapent(struct mapent *me, char *mapent)
+{
+	struct stack *s;
+	char *new;
+
+	if (!me->mapent)
+		return CHE_FAIL;
+
+	if (!mapent)
+		new = NULL;
+	else {
+		new = strdup(mapent);
+		if (!new)
+			return CHE_FAIL;
+	}
+
+	s = malloc(sizeof(struct stack));
+	if (!s) {
+		if (new)
+			free(new);
+		return CHE_FAIL;
+	}
+	memset(s, 0, sizeof(*s));
+
+	s->mapent = me->mapent;
+	s->age = me->age;
+	me->mapent = mapent;
+
+	if (me->stack)
+		s->next = me->stack;
+	me->stack = s;
+
+	return CHE_OK;
+}
+
+/* Restore cache entry mapent to a previously saved mapent, discard current */
+int cache_pop_mapent(struct mapent *me)
+{
+	struct stack *s = me->stack;
+	char *mapent;
+	time_t age;
+
+	if (!s || !s->mapent)
+		return CHE_FAIL;
+
+	mapent = s->mapent;
+	age = s->age;
+	me->stack = s->next;
+	free(s);
+
+	if (age < me->age) {
+		free(mapent);
+		return CHE_OK;
+	}
+
+	if (me->mapent)
+		free(me->mapent);
+	me->mapent = mapent;
+
+	return CHE_OK;
+}
+
 struct mapent_cache *cache_init(struct autofs_point *ap, struct map_source *map)
 {
 	struct mapent_cache *mc;
@@ -578,6 +641,8 @@ int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, c
 	} else
 		me->mapent = NULL;
 
+	me->stack = NULL;
+
 	me->age = age;
 	me->status = 0;
 	me->mc = mc;
@@ -659,8 +724,25 @@ int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *k
 
 	me = cache_lookup_distinct(mc, key);
 	if (me && me->age == age) {
-		if (me == owner || strcmp(me->key, key) == 0)
+		if (me == owner || strcmp(me->key, key) == 0) {
+			char *pent;
+
+			warn(logopt,
+			     "duplcate offset detected for key %s", me->key);
+
+			pent = malloc(strlen(mapent) + 1);
+			if (!pent)
+				warn(logopt,
+				     "map entry not updated: %s", me->mapent);
+			else {
+				if (me->mapent)
+					free(me->mapent);
+				me->mapent = strcpy(pent, mapent);
+				warn(logopt,
+				     "map entry updated with: %s", mapent);
+			}
 			return CHE_DUPLICATE;
+		}
 	}
 
 	ret = cache_update(mc, owner->source, key, mapent, age);
@@ -688,8 +770,14 @@ void cache_update_negative(struct mapent_cache *mc,
 	struct mapent *me;
 	int rv = CHE_OK;
 
+	/* Don't update the wildcard */
+	if (strlen(key) == 1 && *key == '*')
+		return;
+
 	me = cache_lookup_distinct(mc, key);
-	if (!me)
+	if (me)
+		rv = cache_push_mapent(me, NULL);
+	else
 		rv = cache_update(mc, ms, key, NULL, now);
 	if (rv != CHE_FAIL) {
 		me = cache_lookup_distinct(mc, key);
@@ -858,6 +946,7 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 		pred = me;
 		me = me->next;
 		if (strcmp(this, me->key) == 0) {
+			struct stack *s = me->stack;
 			if (me->multi && !list_empty(&me->multi_list)) {
 				ret = CHE_FAIL;
 				goto done;
@@ -872,6 +961,13 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 			free(me->key);
 			if (me->mapent)
 				free(me->mapent);
+			while (s) {
+				struct stack *next = s->next;
+				if (s->mapent)
+					free(s->mapent);
+				free(s);
+				s = next;
+			}
 			free(me);
 			me = pred;
 		}
@@ -882,6 +978,7 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 		goto done;
 
 	if (strcmp(this, me->key) == 0) {
+		struct stack *s = me->stack;
 		if (me->multi && !list_empty(&me->multi_list)) {
 			ret = CHE_FAIL;
 			goto done;
@@ -896,6 +993,13 @@ int cache_delete(struct mapent_cache *mc, const char *key)
 		free(me->key);
 		if (me->mapent)
 			free(me->mapent);
+		while (s) {
+			struct stack *next = s->next;
+			if (s->mapent)
+				free(s->mapent);
+			free(s);
+			s = next;
+		}
 		free(me);
 	}
 done:
@@ -1070,7 +1174,6 @@ struct mapent *cache_enumerate(struct mapent_cache *mc, struct mapent *me)
  * Get each offset from list head under prefix.
  * Maintain traversal current position in pos for subsequent calls. 
  * Return each offset into offset.
- * TODO: length check on offset.
  */
 /* cache must be read locked by caller */
 char *cache_get_offset(const char *prefix, char *offset, int start,
@@ -1099,6 +1202,9 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 			continue;
 
 		if (!strncmp(prefix, offset_start, plen)) {
+			struct mapent *np = NULL;
+			char pe[PATH_MAX + 1];
+
 			/* "/" doesn't count for root offset */
 			if (plen == 1)
 				pstart = &offset_start[plen - 1];
@@ -1111,7 +1217,24 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 
 			/* get next offset */
 			pend = pstart;
-			while (*pend++) ;
+			while (*pend++) {
+				size_t nest_pt_offset;
+
+				if (*pend != '/')
+					continue;
+
+				nest_pt_offset = start + pend - pstart;
+				if (plen > 1)
+					nest_pt_offset += plen;
+				strcpy(pe, this->key);
+				pe[nest_pt_offset] = '\0';
+
+				np = cache_lookup_distinct(this->mc, pe);
+				if (np)
+					break;
+			}
+			if (np)
+				continue;
 			len = pend - pstart - 1;
 			strncpy(offset, pstart, len);
 			offset[len] ='\0';
@@ -1144,7 +1267,9 @@ char *cache_get_offset(const char *prefix, char *offset, int start,
 			break;
 
 		/* compare offset */
-		if (pstart[len] != '/' || strncmp(offset, pstart, len))
+		if (pstart[len] != '/' ||
+		    strlen(pstart) != len ||
+		    strncmp(offset, pstart, len))
 			break;
 
 		*pos = next;
