@@ -17,10 +17,14 @@
 #include <ctype.h>
 #include <string.h>
 
+#include "config.h"
 #include "list.h"
 #include "defaults.h"
+#ifdef WITH_LDAP
 #include "lookup_ldap.h"
+#endif
 #include "log.h"
+#include "automount.h"
 
 #define DEFAULTS_CONFIG_FILE		AUTOFS_CONF_DIR "/autofs"
 #define MAX_LINE_LEN			256
@@ -28,9 +32,11 @@
 #define ENV_NAME_MASTER_MAP		"MASTER_MAP_NAME"
 
 #define ENV_NAME_TIMEOUT		"TIMEOUT"
+#define ENV_NAME_MASTER_WAIT		"MASTER_WAIT"
 #define ENV_NAME_NEGATIVE_TIMEOUT	"NEGATIVE_TIMEOUT"
 #define ENV_NAME_BROWSE_MODE		"BROWSE_MODE"
 #define ENV_NAME_LOGGING		"LOGGING"
+#define ENV_NAME_FORCE_STD_PROG_MAP_ENV	"FORCE_STANDARD_PROGRAM_MAP_ENV"
 
 #define LDAP_URI			"LDAP_URI"
 #define ENV_LDAP_TIMEOUT		"LDAP_TIMEOUT"
@@ -44,9 +50,14 @@
 #define ENV_NAME_ENTRY_ATTR		"ENTRY_ATTRIBUTE"
 #define ENV_NAME_VALUE_ATTR		"VALUE_ATTRIBUTE"
 
+#define ENV_MOUNT_NFS_DEFAULT_PROTOCOL	"MOUNT_NFS_DEFAULT_PROTOCOL"
 #define ENV_APPEND_OPTIONS		"APPEND_OPTIONS"
+#define ENV_MOUNT_WAIT			"MOUNT_WAIT"
+#define ENV_UMOUNT_WAIT			"UMOUNT_WAIT"
 #define ENV_AUTH_CONF_FILE		"AUTH_CONF_FILE"
 
+#define ENV_MAP_HASH_TABLE_SIZE		"MAP_HASH_TABLE_SIZE"
+
 static const char *default_master_map_name = DEFAULT_MASTER_MAP_NAME;
 static const char *default_auth_conf_file  = DEFAULT_AUTH_CONF_FILE;
 
@@ -59,7 +70,7 @@ static char *get_env_string(const char *name)
 		return NULL;
 
 	res = strdup(val);
-	if (!val)
+	if (!res)
 		return NULL;
 
 	return res;
@@ -158,7 +169,7 @@ static int parse_line(char *line, char **res, char **value)
 	while (*key && *key == ' ')
 		key++;
 
-	if (!key)
+	if (!*key)
 		return 0;
 
 	if (!(val = strchr(key, '=')))
@@ -191,6 +202,7 @@ static int parse_line(char *line, char **res, char **value)
 	return 1;
 }
 
+#ifdef WITH_LDAP
 void defaults_free_uris(struct list_head *list)
 {
 	struct list_head *next;
@@ -217,7 +229,7 @@ void defaults_free_uris(struct list_head *list)
 static unsigned int add_uris(char *value, struct list_head *list)
 {
 	char *str, *tok, *ptr = NULL;
-	size_t len = strlen(value);
+	size_t len = strlen(value) + 1;
 
 	str = alloca(len);
 	if (!str)
@@ -254,7 +266,7 @@ struct list_head *defaults_get_uris(void)
 	char *res;
 	struct list_head *list;
 
-	f = fopen(DEFAULTS_CONFIG_FILE, "r");
+	f = open_fopen_r(DEFAULTS_CONFIG_FILE);
 	if (!f)
 		return NULL;
 
@@ -284,162 +296,6 @@ struct list_head *defaults_get_uris(void)
 	return list;
 }
 
-/*
- * Read config env variables and check they have been set.
- *
- * This simple minded routine assumes the config file
- * is valid bourne shell script without spaces around "="
- * and that it has valid values.
- */
-unsigned int defaults_read_config(unsigned int to_syslog)
-{
-	FILE *f;
-	char buf[MAX_LINE_LEN];
-	char *res;
-
-	f = fopen(DEFAULTS_CONFIG_FILE, "r");
-	if (!f)
-		return 0;
-
-	while ((res = fgets(buf, MAX_LINE_LEN, f))) {
-		char *key, *value;
-
-		if (!parse_line(res, &key, &value))
-			continue;
-
-		if (check_set_config_value(key, ENV_NAME_MASTER_MAP, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_TIMEOUT, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_NEGATIVE_TIMEOUT, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_BROWSE_MODE, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_LOGGING, value, to_syslog) ||
-		    check_set_config_value(key, ENV_LDAP_TIMEOUT, value, to_syslog) ||
-		    check_set_config_value(key, ENV_LDAP_NETWORK_TIMEOUT, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_MAP_OBJ_CLASS, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_ENTRY_OBJ_CLASS, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_MAP_ATTR, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_ENTRY_ATTR, value, to_syslog) ||
-		    check_set_config_value(key, ENV_NAME_VALUE_ATTR, value, to_syslog) ||
-		    check_set_config_value(key, ENV_APPEND_OPTIONS, value, to_syslog) ||
-		    check_set_config_value(key, ENV_AUTH_CONF_FILE, value, to_syslog))
-			;
-	}
-
-	if (!feof(f) || ferror(f)) {
-		if (!to_syslog) {
-			fprintf(stderr,
-				"fgets returned error %d while reading %s\n",
-				ferror(f), DEFAULTS_CONFIG_FILE);
-		} else {
-			logmsg("fgets returned error %d while reading %s",
-			      ferror(f), DEFAULTS_CONFIG_FILE);
-		}
-		fclose(f);
-		return 0;
-	}
-
-	fclose(f);
-	return 1;
-}
-
-const char *defaults_get_master_map(void)
-{
-	char *master;
-
-	master = get_env_string(ENV_NAME_MASTER_MAP);
-	if (!master)
-		return strdup(default_master_map_name);
-
-	return (const char *) master;
-}
-
-int defaults_master_set(void)
-{
-	char *val = getenv(ENV_NAME_MASTER_MAP);
-	if (!val)
-		return 0;
-
-	return 1;
-}
-
-unsigned int defaults_get_timeout(void)
-{
-	long timeout;
-
-	timeout = get_env_number(ENV_NAME_TIMEOUT);
-	if (timeout < 0)
-		timeout = DEFAULT_TIMEOUT;
-
-	return (unsigned int) timeout;
-}
-
-unsigned int defaults_get_negative_timeout(void)
-{
-	long n_timeout;
-
-	n_timeout = get_env_number(ENV_NAME_NEGATIVE_TIMEOUT);
-	if (n_timeout <= 0)
-		n_timeout = DEFAULT_NEGATIVE_TIMEOUT;
-
-	return (unsigned int) n_timeout;
-}
-
-unsigned int defaults_get_browse_mode(void)
-{
-	int res;
-
-	res = get_env_yesno(ENV_NAME_BROWSE_MODE);
-	if (res < 0)
-		res = DEFAULT_BROWSE_MODE;
-
-	return res;
-}
-
-unsigned int defaults_get_logging(void)
-{
-	char *res;
-	unsigned int logging = DEFAULT_LOGGING;
-
-	res = get_env_string(ENV_NAME_LOGGING);
-	if (!res)
-		return logging;
-
-	if (!strcasecmp(res, "none"))
-		logging = DEFAULT_LOGGING;
-	else {
-		if (!strcasecmp(res, "verbose"))
-			logging |= LOGOPT_VERBOSE;
-
-		if (!strcasecmp(res, "debug"))
-			logging |= LOGOPT_DEBUG;
-	}
-
-	free(res);
-
-	return logging;
-}
-
-unsigned int defaults_get_ldap_timeout(void)
-{
-	int res;
-
-	res = get_env_number(ENV_LDAP_TIMEOUT);
-	if (res < 0)
-		res = DEFAULT_LDAP_TIMEOUT;
-
-	return res;
-}
-
-unsigned int defaults_get_ldap_network_timeout(void)
-{
-	int res;
-
-	res = get_env_number(ENV_LDAP_NETWORK_TIMEOUT);
-	if (res < 0)
-		res = DEFAULT_LDAP_NETWORK_TIMEOUT;
-
-	return res;
-}
-
 struct ldap_schema *defaults_get_default_schema(void)
 {
 	struct ldap_schema *schema;
@@ -524,7 +380,6 @@ void defaults_free_searchdns(struct ldap_searchdn *sdn)
 	struct ldap_searchdn *this = sdn;
 	struct ldap_searchdn *next;
 
-	next = this;
 	while (this) {
 		next = this->next;
 		free(this->basedn);
@@ -542,7 +397,7 @@ struct ldap_searchdn *defaults_get_searchdns(void)
 	char *res;
 	struct ldap_searchdn *sdn, *last;
 
-	f = fopen(DEFAULTS_CONFIG_FILE, "r");
+	f = open_fopen_r(DEFAULTS_CONFIG_FILE);
 	if (!f)
 		return NULL;
 
@@ -559,6 +414,7 @@ struct ldap_searchdn *defaults_get_searchdns(void)
 
 			if (!new) {
 				defaults_free_searchdns(sdn);
+				fclose(f);
 				return NULL;
 			}
 
@@ -635,6 +491,202 @@ struct ldap_schema *defaults_get_schema(void)
 
 	return schema;
 }
+#endif
+
+/*
+ * Read config env variables and check they have been set.
+ *
+ * This simple minded routine assumes the config file
+ * is valid bourne shell script without spaces around "="
+ * and that it has valid values.
+ */
+unsigned int defaults_read_config(unsigned int to_syslog)
+{
+	FILE *f;
+	char buf[MAX_LINE_LEN];
+	char *res;
+
+	f = open_fopen_r(DEFAULTS_CONFIG_FILE);
+	if (!f)
+		return 0;
+
+	while ((res = fgets(buf, MAX_LINE_LEN, f))) {
+		char *key, *value;
+
+		if (!parse_line(res, &key, &value))
+			continue;
+
+		if (check_set_config_value(key, ENV_NAME_MASTER_MAP, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_TIMEOUT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_MASTER_WAIT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_NEGATIVE_TIMEOUT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_BROWSE_MODE, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_LOGGING, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_FORCE_STD_PROG_MAP_ENV, value, to_syslog) ||
+		    check_set_config_value(key, ENV_LDAP_TIMEOUT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_LDAP_NETWORK_TIMEOUT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_MAP_OBJ_CLASS, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_ENTRY_OBJ_CLASS, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_MAP_ATTR, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_ENTRY_ATTR, value, to_syslog) ||
+		    check_set_config_value(key, ENV_NAME_VALUE_ATTR, value, to_syslog) ||
+		    check_set_config_value(key, ENV_APPEND_OPTIONS, value, to_syslog) ||
+		    check_set_config_value(key, ENV_MOUNT_WAIT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_UMOUNT_WAIT, value, to_syslog) ||
+		    check_set_config_value(key, ENV_AUTH_CONF_FILE, value, to_syslog) ||
+		    check_set_config_value(key, ENV_MAP_HASH_TABLE_SIZE, value, to_syslog) ||
+		    check_set_config_value(key, ENV_MOUNT_NFS_DEFAULT_PROTOCOL, value, to_syslog))
+			;
+	}
+
+	if (!feof(f) || ferror(f)) {
+		if (!to_syslog) {
+			fprintf(stderr,
+				"fgets returned error %d while reading %s\n",
+				ferror(f), DEFAULTS_CONFIG_FILE);
+		} else {
+			logmsg("fgets returned error %d while reading %s",
+			      ferror(f), DEFAULTS_CONFIG_FILE);
+		}
+		fclose(f);
+		return 0;
+	}
+
+	fclose(f);
+	return 1;
+}
+
+const char *defaults_get_master_map(void)
+{
+	char *master;
+
+	master = get_env_string(ENV_NAME_MASTER_MAP);
+	if (!master)
+		return strdup(default_master_map_name);
+
+	return (const char *) master;
+}
+
+int defaults_master_set(void)
+{
+	char *val = getenv(ENV_NAME_MASTER_MAP);
+	if (!val)
+		return 0;
+
+	return 1;
+}
+
+unsigned int defaults_get_timeout(void)
+{
+	long timeout;
+
+	timeout = get_env_number(ENV_NAME_TIMEOUT);
+	if (timeout < 0)
+		timeout = DEFAULT_TIMEOUT;
+
+	return (unsigned int) timeout;
+}
+
+int defaults_get_master_wait(void)
+{
+	long wait;
+
+	wait = get_env_number(ENV_NAME_MASTER_WAIT);
+	if (wait < 0)
+		wait = DEFAULT_MASTER_WAIT;
+
+	return (int) wait;
+}
+
+unsigned int defaults_get_negative_timeout(void)
+{
+	long n_timeout;
+
+	n_timeout = get_env_number(ENV_NAME_NEGATIVE_TIMEOUT);
+	if (n_timeout <= 0)
+		n_timeout = DEFAULT_NEGATIVE_TIMEOUT;
+
+	return (unsigned int) n_timeout;
+}
+
+unsigned int defaults_get_browse_mode(void)
+{
+	int res;
+
+	res = get_env_yesno(ENV_NAME_BROWSE_MODE);
+	if (res < 0)
+		res = DEFAULT_BROWSE_MODE;
+
+	return res;
+}
+
+unsigned int defaults_get_logging(void)
+{
+	char *res;
+	unsigned int logging = DEFAULT_LOGGING;
+
+	res = get_env_string(ENV_NAME_LOGGING);
+	if (!res)
+		return logging;
+
+	if (!strcasecmp(res, "none"))
+		logging = DEFAULT_LOGGING;
+	else {
+		if (!strcasecmp(res, "verbose"))
+			logging |= LOGOPT_VERBOSE;
+
+		if (!strcasecmp(res, "debug"))
+			logging |= LOGOPT_DEBUG;
+	}
+
+	free(res);
+
+	return logging;
+}
+
+unsigned int defaults_force_std_prog_map_env(void)
+{
+	int res;
+
+	res = get_env_yesno(ENV_NAME_FORCE_STD_PROG_MAP_ENV);
+	if (res < 0)
+		res = DEFAULT_FORCE_STD_PROG_MAP_ENV;
+
+	return res;
+}
+
+unsigned int defaults_get_ldap_timeout(void)
+{
+	int res;
+
+	res = get_env_number(ENV_LDAP_TIMEOUT);
+	if (res < 0)
+		res = DEFAULT_LDAP_TIMEOUT;
+
+	return res;
+}
+
+unsigned int defaults_get_ldap_network_timeout(void)
+{
+	int res;
+
+	res = get_env_number(ENV_LDAP_NETWORK_TIMEOUT);
+	if (res < 0)
+		res = DEFAULT_LDAP_NETWORK_TIMEOUT;
+
+	return res;
+}
+
+unsigned int defaults_get_mount_nfs_default_proto(void)
+{
+	long proto;
+
+	proto = get_env_number(ENV_MOUNT_NFS_DEFAULT_PROTOCOL);
+	if (proto < 2 || proto > 4)
+		proto = DEFAULT_NFS_MOUNT_PROTOCOL;
+
+	return (unsigned int) proto;
+}
 
 unsigned int defaults_get_append_options(void)
 {
@@ -647,6 +699,28 @@ unsigned int defaults_get_append_options(void)
 	return res;
 }
 
+unsigned int defaults_get_mount_wait(void)
+{
+	long wait;
+
+	wait = get_env_number(ENV_MOUNT_WAIT);
+	if (wait < 0)
+		wait = DEFAULT_MOUNT_WAIT;
+
+	return (unsigned int) wait;
+}
+
+unsigned int defaults_get_umount_wait(void)
+{
+	long wait;
+
+	wait = get_env_number(ENV_UMOUNT_WAIT);
+	if (wait < 0)
+		wait = DEFAULT_UMOUNT_WAIT;
+
+	return (unsigned int) wait;
+}
+
 const char *defaults_get_auth_conf_file(void)
 {
 	char *cf;
@@ -658,3 +732,14 @@ const char *defaults_get_auth_conf_file(void)
 	return (const char *) cf;
 }
 
+unsigned int defaults_get_map_hash_table_size(void)
+{
+	long size;
+
+	size = get_env_number(ENV_MAP_HASH_TABLE_SIZE);
+	if (size < 0)
+		size = DEFAULT_MAP_HASH_TABLE_SIZE;
+
+	return (unsigned int) size;
+}
+
