@@ -18,8 +18,6 @@
 #include <stdio.h>
 #include <malloc.h>
 #include <netdb.h>
-#include <fcntl.h>
-#include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -39,14 +37,25 @@
 
 #define MODPREFIX "parse(sun): "
 
-#define MOUNT_MOVE_NONE		0x00
-#define MOUNT_MOVE_AUTOFS	0x01
-#define MOUNT_MOVE_OTHER	0x02
-
 int parse_version = AUTOFS_PARSE_VERSION;	/* Required by protocol */
 
 static struct mount_mod *mount_nfs = NULL;
 static int init_ctr = 0;
+static pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+static void instance_mutex_lock(void)
+{
+	int status = pthread_mutex_lock(&instance_mutex);
+	if (status)
+		fatal(status);
+}
+
+static void instance_mutex_unlock(void)
+{
+	int status = pthread_mutex_unlock(&instance_mutex);
+	if (status)
+		fatal(status);
+}
 
 extern const char *global_options;
 
@@ -89,44 +98,6 @@ static void kill_context(struct parse_context *ctxt)
 	free(ctxt);
 }
 
-static struct substvar *addstdenv(struct substvar *sv)
-{
-	struct substvar *list = sv;
-	struct thread_stdenv_vars *tsv;
-	char numbuf[16];
-
-	tsv = pthread_getspecific(key_thread_stdenv_vars);
-	if (tsv) {
-		int ret;
-		long num;
-
-		num = (long) tsv->uid;
-		ret = sprintf(numbuf, "%ld", num);
-		if (ret > 0)
-			list = macro_addvar(list, "UID", 3, numbuf);
-		num = (long) tsv->gid;
-		ret = sprintf(numbuf, "%ld", num);
-		if (ret > 0)
-			list = macro_addvar(list, "GID", 3, numbuf);
-		list = macro_addvar(list, "USER", 4, tsv->user);
-		list = macro_addvar(list, "GROUP", 5, tsv->group);
-		list = macro_addvar(list, "HOME", 4, tsv->home);
-	}
-	return list;
-}
-
-static struct substvar *removestdenv(struct substvar *sv)
-{
-	struct substvar *list = sv;
-
-	list = macro_removevar(list, "UID", 3);
-	list = macro_removevar(list, "USER", 4);
-	list = macro_removevar(list, "HOME", 4);
-	list = macro_removevar(list, "GID", 3);
-	list = macro_removevar(list, "GROUP", 5);
-	return list;
-}
-
 /* 
  * $- and &-expand a Sun-style map entry and return the length of the entry.
  * If "dst" is NULL, just count the length.
@@ -147,30 +118,24 @@ int expandsunent(const char *src, char *dst, const char *key,
 		case '&':
 			l = strlen(key);
 			/*
-			 * In order to ensure that any spaces in the key
-			 * re preserved, we need to escape them here.
+			 * In order to ensure that any isspace() characters
+			 * in the key are preserved, we need to escape them
+			 * here.
 			 */
-			if (strchr(key, ' ')) {
-				const char *keyp = key;
-				while (*keyp) {
-					if (isspace(*keyp)) {
-						if (dst) {
-							*dst++ = '\\';
-							*dst++ = *keyp++;
-						} else
-							keyp++;
-						l++;
-					} else {
-						if (dst)
-							*dst++ = *keyp++;
-						else
-							keyp++;
-					}
-				}
-			} else {
-				if (dst) {
-					strcpy(dst, key);
-					dst += l;
+			const char *keyp = key;
+			while (*keyp) {
+				if (isspace(*keyp)) {
+					if (dst) {
+						*dst++ = '\\';
+						*dst++ = *keyp++;
+					} else
+						keyp++;
+					l++;
+				} else {
+					if (dst)
+						*dst++ = *keyp++;
+					else
+						keyp++;
 				}
 			}
 			len += l;
@@ -247,7 +212,9 @@ int expandsunent(const char *src, char *dst, const char *key,
 				*(dst++) = 
 				  (seen_colons && slashify_colons) ? '/' : ':';
 			len++;
-			seen_colons = 1;
+			/* Were looking for the colon preceeding a path */
+			if (*src == '/')
+				seen_colons = 1;
 			break;
 
 		default:
@@ -275,11 +242,6 @@ int parse_init(int argc, const char *const *argv, void **context)
 	int i, bval;
 	unsigned int append_options;
 
-	/* Get processor information for predefined escapes */
-
-	if (!init_ctr)
-		macro_init();
-
 	/* Set up context and escape chain */
 
 	if (!(ctxt = (struct parse_context *) malloc(sizeof(struct parse_context)))) {
@@ -381,15 +343,17 @@ int parse_init(int argc, const char *const *argv, void **context)
 			if (ctxt->optstr) {
 				noptstr =
 				    (char *) realloc(ctxt->optstr, optlen + len + 2);
-				if (!noptstr)
-					break;
-				noptstr[optlen] = ',';
-				strcpy(noptstr + optlen + 1, argv[i] + offset);
-				optlen += len + 1;
+				if (noptstr) {
+					noptstr[optlen] = ',';
+					strcpy(noptstr + optlen + 1, argv[i] + offset);
+					optlen += len + 1;
+				}
 			} else {
 				noptstr = (char *) malloc(len + 1);
-				strcpy(noptstr, argv[i] + offset);
-				optlen = len;
+				if (noptstr) {
+					strcpy(noptstr, argv[i] + offset);
+					optlen = len;
+				}
 			}
 			if (!noptstr) {
 				char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
@@ -432,19 +396,21 @@ options_done:
 
 	/* We only need this once.  NFS mounts are so common that we cache
 	   this module. */
-	if (!mount_nfs) {
+	instance_mutex_lock();
+	if (mount_nfs)
+		init_ctr++;
+	else {
 		if ((mount_nfs = open_mount("nfs", MODPREFIX))) {
 			init_ctr++;
-			return 0;
 		} else {
 			kill_context(ctxt);
 			*context = NULL;
+			instance_mutex_unlock();
 			return 1;
 		}
-	} else {
-		init_ctr++;
-		return 0;
 	}
+	instance_mutex_unlock();
+	return 0;
 }
 
 static const char *parse_options(const char *str, char **ret, unsigned int logopt)
@@ -506,6 +472,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 {
 	char *fstype = "nfs";	/* Default filesystem type */
 	int nonstrict = 1;
+	int use_weight_only = ap->flags & MOUNT_FLAG_USE_WEIGHT_ONLY;
 	int rv, cur_state;
 	char *mountpoint;
 	char *what;
@@ -535,25 +502,30 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 			while (*comma != '\0' && *comma != ',')
 				comma++;
 
-			if (strncmp("fstype=", cp, 7) == 0) {
+			if (_strncmp("fstype=", cp, 7) == 0) {
 				int typelen = comma - (cp + 7);
 				fstype = alloca(typelen + 1);
 				memcpy(fstype, cp + 7, typelen);
 				fstype[typelen] = '\0';
-			} else if (strncmp("nonstrict", cp, 9) == 0) {
+			} else if (_strncmp("nonstrict", cp, 9) == 0) {
 				nonstrict = 1;
-			} else if (strncmp("strict", cp, 6) == 0) {
+			} else if (_strncmp("strict", cp, 6) == 0 &&
+				   comma - cp == 6) {
 				nonstrict = 0;
-			} else if (strncmp("nobrowse", cp, 8) == 0 ||
-				   strncmp("browse", cp, 6) == 0 ||
-				   strncmp("timeout=", cp, 8) == 0) {
+			} else if (_strncmp("nobrowse", cp, 8) == 0 ||
+				   _strncmp("browse", cp, 6) == 0 ||
+				   _strncmp("timeout=", cp, 8) == 0) {
 				if (strcmp(fstype, "autofs") == 0 ||
 				    strstr(cp, "fstype=autofs")) {
 					memcpy(np, cp, comma - cp + 1);
 					np += comma - cp + 1;
 				}
-			} else if (strncmp("bg", cp, 2) == 0 ||
-				   strncmp("nofg", cp, 4) == 0) {
+			} else if (_strncmp("no-use-weight-only", cp, 18) == 0) {
+				use_weight_only = -1;
+			} else if (_strncmp("use-weight-only", cp, 15) == 0) {
+				use_weight_only = MOUNT_FLAG_USE_WEIGHT_ONLY;
+			} else if (_strncmp("bg", cp, 2) == 0 ||
+				   _strncmp("nofg", cp, 4) == 0) {
 				continue;
 			} else {
 				memcpy(np, cp, comma - cp + 1);
@@ -570,11 +542,10 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		options = noptions;
 	}
 
-
 	if (!strcmp(fstype, "autofs") && ctxt->macros) {
 		char *noptions = NULL;
 
-		if (!options) {
+		if (!options || *options == '\0') {
 			noptions = alloca(strlen(ctxt->macros) + 1);
 			*noptions = '\0';
 		} else {
@@ -587,7 +558,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 			}
 		}
 
-		if (noptions) {
+		if (noptions && *noptions != '\0') {
 			strcat(noptions, ctxt->macros);
 			options = noptions;
 		} else {
@@ -601,13 +572,14 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 
 	type = ap->entry->maps->type;
 	if (type && !strcmp(type, "hosts")) {
-		if (options) {
+		if (options && *options != '\0') {
 			int len = strlen(options);
 			int suid = strstr(options, "suid") ? 0 : 7;
 			int dev = strstr(options, "dev") ? 0 : 6;
+			int nointr = strstr(options, "nointr") ? 0 : 5;
 
-			if (suid || dev) {
-				char *tmp = alloca(len + suid + dev + 1);
+			if (suid || dev || nointr) {
+				char *tmp = alloca(len + suid + dev + nointr + 1);
 				if (!tmp) {
 					error(ap->logopt, MODPREFIX
 					      "alloca failed for options");
@@ -621,10 +593,12 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					strcat(tmp, ",nosuid");
 				if (dev)
 					strcat(tmp, ",nodev");
+				if (nointr)
+					strcat(tmp, ",intr");
 				options = tmp;
 			}
 		} else {
-			char *tmp = alloca(13);
+			char *tmp = alloca(18);
 			if (!tmp) {
 				error(ap->logopt,
 				      MODPREFIX "alloca failed for options");
@@ -632,7 +606,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					return -1;
 				return 1;
 			}
-			strcpy(tmp, "nosuid,nodev");
+			strcpy(tmp, "nosuid,nodev,intr");
 			options = tmp;
 		}
 	}
@@ -643,6 +617,30 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		memcpy(what, loc, loclen);
 		what[loclen] = '\0';
 
+		/* Add back "[no-]use-weight-only" for NFS mounts only */
+		if (use_weight_only) {
+			char *tmp;
+			int len;
+
+			if (options && *options != '\0') {
+				len = strlen(options) + 19;
+				tmp = alloca(len);
+				strcpy(tmp, options);
+				strcat(tmp, ",");
+				if (use_weight_only == MOUNT_FLAG_USE_WEIGHT_ONLY)
+					strcat(tmp, "use-weight-only");
+				else
+					strcat(tmp, "no-use-weight-only");
+			} else {
+				tmp = alloca(19);
+				if (use_weight_only == MOUNT_FLAG_USE_WEIGHT_ONLY)
+					strcpy(tmp, "use-weight-only");
+				else
+					strcpy(tmp, "no-use-weight-only");
+			}
+			options = tmp;
+		}
+
 		debug(ap->logopt, MODPREFIX
 		      "mounting root %s, mountpoint %s, "
 		      "what %s, fstype %s, options %s",
@@ -651,14 +649,18 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		rv = mount_nfs->mount_mount(ap, root, mountpoint, strlen(mountpoint),
 					    what, fstype, options, mount_nfs->context);
 	} else {
-		what = alloca(loclen + 1);
-		if (*loc == ':') {
-			loclen--;
-			memcpy(what, loc + 1, loclen);
-			what[loclen] = '\0';
-		} else {
-			memcpy(what, loc, loclen);
-			what[loclen] = '\0';
+		if (!loclen)
+			what = NULL;
+		else {
+			what = alloca(loclen + 1);
+			if (*loc == ':') {
+				loclen--;
+				memcpy(what, loc + 1, loclen);
+				what[loclen] = '\0';
+			} else {
+				memcpy(what, loc, loclen);
+				what[loclen] = '\0';
+			}
 		}
 
 		debug(ap->logopt, MODPREFIX
@@ -737,10 +739,10 @@ static int check_is_multi(const char *mapent)
 }
 
 static int
-add_offset_entry(struct autofs_point *ap, const char *name,
-		 const char *m_root, int m_root_len,
-		 const char *path, const char *myoptions, const char *loc,
-		 time_t age)
+update_offset_entry(struct autofs_point *ap, const char *name,
+		    const char *m_root, int m_root_len,
+		    const char *path, const char *myoptions, const char *loc,
+		    time_t age)
 {
 	struct map_source *source;
 	struct mapent_cache *mc;
@@ -755,7 +757,10 @@ add_offset_entry(struct autofs_point *ap, const char *name,
 
 	mc = source->mc;
 
-	if (!*path || !*loc) {
+	memset(m_mapent, 0, MAPENT_MAX_LEN + 1);
+
+	/* Internal hosts map may have loc == NULL */
+	if (!*path) {
 		error(ap->logopt,
 		      MODPREFIX "syntax error in offset %s -> %s", path, loc);
 		return CHE_FAIL;
@@ -780,7 +785,7 @@ add_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions)
 		m_options_len = strlen(myoptions) + 2;
 
-	m_mapent_len = strlen(loc);
+	m_mapent_len = loc ? strlen(loc) : 0;
 	if (m_mapent_len + m_options_len > MAPENT_MAX_LEN) {
 		error(ap->logopt, MODPREFIX "multi mount mapent too long");
 		return CHE_FAIL;
@@ -789,23 +794,34 @@ add_offset_entry(struct autofs_point *ap, const char *name,
 	if (*myoptions) {
 		strcpy(m_mapent, "-");
 		strcat(m_mapent, myoptions);
-		strcat(m_mapent, " ");
-		strcat(m_mapent, loc);
-	} else
-		strcpy(m_mapent, loc);
+		if (loc) {
+			strcat(m_mapent, " ");
+			if (loc)
+				strcat(m_mapent, loc);
+		}
+	} else {
+		if (loc)
+			strcpy(m_mapent, loc);
+	}
 
-	ret = cache_add_offset(mc, name, m_key, m_mapent, age);
-	if (ret == CHE_OK)
-		debug(ap->logopt, MODPREFIX
-		      "added multi-mount offset %s -> %s", path, m_mapent);
-	else
+	ret = cache_update_offset(mc, name, m_key, m_mapent, age);
+	if (ret == CHE_DUPLICATE) {
 		warn(ap->logopt, MODPREFIX
-		      "syntax error or duplicate offset %s -> %s", path, loc);
+		     "syntax error or duplicate offset %s -> %s", path, loc);
+		ret = CHE_OK;
+	} else if (ret == CHE_FAIL)
+		debug(ap->logopt, MODPREFIX
+		      "failed to update multi-mount offset %s -> %s", path, m_mapent);
+	else {
+		ret = CHE_OK;
+		debug(ap->logopt, MODPREFIX
+		      "updated multi-mount offset %s -> %s", path, m_mapent);
+	}
 
 	return ret;
 }
 
-static int validate_location(char *loc)
+static int validate_location(unsigned int logopt, char *loc)
 {
 	char *ptr = loc;
 
@@ -814,27 +830,53 @@ static int validate_location(char *loc)
 		return 1;
 
 	/*
-	 * If a ':' is present now it must be a host name, except
+	 * If a ':/' is present now it must be a host name, except
 	 * for those special file systems like sshfs which use "#"
-	 * and "@" in the host name part.
+	 * and "@" in the host name part and ipv6 addresses that
+	 * have ":", "[" and "]".
 	 */
-	if (check_colon(ptr)) {
-		while (*ptr && *ptr != ':') {
+	if (!check_colon(ptr)) {
+		char *esc;
+		/*
+		 * Don't forget cases where a colon is present but
+		 * not followed by a "/" or, if there is no colon at
+		 * all, we don't know if it is actually invalid since
+		 * it may be a map name by itself, for example.
+		 */
+		if (!strchr(ptr, ':') ||
+		    ((esc = strchr(ptr, '\\')) && *(esc + 1) == ':') ||
+		    !strncmp(ptr, "file:", 5) || !strncmp(ptr, "yp:", 3) ||
+		    !strncmp(ptr, "nis:", 4) || !strncmp(ptr, "nisplus:", 8) ||
+		    !strncmp(ptr, "ldap:", 5) || !strncmp(ptr, "ldaps:", 6) ||
+		    !strncmp(ptr, "sss:", 4) || !strncmp(ptr, "dir:", 4))
+			return 1;
+		error(logopt,
+		      "expected colon delimeter not found in location %s",
+		      loc);
+		return 0;
+	} else {
+		while (*ptr && strncmp(ptr, ":/", 2)) {
 			if (!(isalnum(*ptr) ||
 			    *ptr == '-' || *ptr == '.' || *ptr == '_' ||
 			    *ptr == ',' || *ptr == '(' || *ptr == ')' ||
-			    *ptr == '#' || *ptr == '@'))
+			    *ptr == '#' || *ptr == '@' || *ptr == ':' ||
+			    *ptr == '[' || *ptr == ']' || *ptr == '%')) {
+				error(logopt, "invalid character \"%c\" "
+				      "found in location %s", *ptr, loc);
 				return 0;
+			}
 			ptr++;
 		}
 
-		if (*ptr && *ptr == ':')
+		if (*ptr && !strncmp(ptr, ":/", 2))
 			ptr++;
 	}
 
 	/* Must always be something following */
-	if (!*ptr)
+	if (!*ptr) {
+		error(logopt, "invalid location %s", loc);
 		return 0;
+	}
 
 	return 1;
 }
@@ -888,20 +930,26 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	l = chunklen(p, check_colon(p));
 	loc = dequote(p, l, logopt);
 	if (!loc) {
-		warn(logopt, MODPREFIX "possible missing location");
-		free(myoptions);
-		return 0;
+		if (strstr(myoptions, "fstype=autofs") &&
+		    strstr(myoptions, "hosts")) {
+			warn(logopt, MODPREFIX "possible missing location");
+			free(myoptions);
+			return 0;
+		}
+		*options = myoptions;
+		*location = NULL;
+		return (p - ent);
 	}
 
 	/* Location can't begin with a '/' */
 	if (*p == '/') {
 		warn(logopt, MODPREFIX "error location begins with \"/\"");
 		free(myoptions);
+		free(loc);
 		return 0;
 	}
 
-	if (!validate_location(loc)) {
-		warn(logopt, MODPREFIX "invalid location %s", loc);
+	if (!validate_location(logopt, loc)) {
 		free(myoptions);
 		free(loc);
 		return 0;
@@ -918,10 +966,15 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		l = chunklen(p, check_colon(p));
 		ent_chunk = dequote(p, l, logopt);
 		if (!ent_chunk) {
-			warn(logopt, MODPREFIX "null location or out of memory");
-			free(myoptions);
-			free(loc);
-			return 0;
+			if (strstr(myoptions, "fstype=autofs") &&
+			    strstr(myoptions, "hosts")) {
+				warn(logopt, MODPREFIX
+				     "null location or out of memory");
+				free(myoptions);
+				free(loc);
+				return 0;
+			}
+			goto next;
 		}
 
 		/* Location can't begin with a '/' */
@@ -934,9 +987,7 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 			return 0;
 		}
 
-		if (!validate_location(ent_chunk)) {
-			warn(logopt,
-			      MODPREFIX "invalid location %s", ent_chunk);
+		if (!validate_location(logopt, ent_chunk)) {
 			free(ent_chunk);
 			free(myoptions);
 			free(loc);
@@ -959,7 +1010,7 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 		strcat(loc, ent_chunk);
 
 		free(ent_chunk);
-
+next:
 		p += l;
 		p = skipspace(p);
 	}
@@ -970,84 +1021,6 @@ static int parse_mapent(const char *ent, char *g_options, char **options, char *
 	return (p - ent);
 }
 
-static int move_mount(struct autofs_point *ap,
-		      const char *mm_tmp_root, const char *mm_root,
-		      unsigned int move)
-{
-	char buf[MAX_ERR_BUF];
-	int err;
-
-	if (move == MOUNT_MOVE_NONE)
-		return 1;
-
-	err = mkdir_path(mm_root, 0555);
-	if (err < 0 && errno != EEXIST) {
-		error(ap->logopt,
-		      "failed to create move target mount point %s", mm_root);
-		return 0;
-	}
-
-	if (move == MOUNT_MOVE_AUTOFS)
-		err = mount(mm_tmp_root, mm_root, NULL, MS_MOVE, NULL);
-	else
-		err = spawn_mount(ap->logopt,
-				  "--move", mm_tmp_root, mm_root, NULL);
-	if (err) {
-		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt,
-		      "failed to move mount from %s to %s: %s",
-		      mm_tmp_root, mm_root, estr);
-		return 0;
-	}
-
-	debug(ap->logopt,
-	      "moved mount tree from %s to %s", mm_tmp_root, mm_root);
-
-	return 1;
-}
-
-static void cleanup_multi_root(struct autofs_point *ap, const char *root,
-					 const char *path, unsigned int move)
-{
-	if (move == MOUNT_MOVE_NONE)
-		return;
-
-	if (move == MOUNT_MOVE_OTHER)
-		spawn_umount(ap->logopt, root, NULL);
-	else {
-		struct ioctl_ops *ops = get_ioctl_ops();
-		struct autofs_point *submount;
-
-		mounts_mutex_lock(ap);
-		submount = __master_find_submount(ap, path);
-		if (!submount) {
-			mounts_mutex_unlock(ap);
-			return;
-		}
-
-		alarm_delete(submount);
-		st_remove_tasks(submount);
-		st_wait_state(submount, ST_READY);
-
-		submount->parent->submnt_count--;
-		list_del_init(&submount->mounts);
-
-		ops->catatonic(submount->logopt, submount->ioctlfd);
-
-		mounts_mutex_unlock(ap);
-
-		if (submount->thid) {
-			pthread_cancel(submount->thid);
-			close_mount_fds(submount);
-			umount(root);
-			destroy_logpri_fifo(submount);
-			master_free_mapent_sources(submount->entry, 1);
-			master_free_mapent(ap->entry);
-		}
-	}
-	return;
-}
-
 static void cleanup_multi_triggers(struct autofs_point *ap,
 			    struct mapent *me, const char *root, int start,
 			    const char *base)
@@ -1087,63 +1060,20 @@ static void cleanup_multi_triggers(struct autofs_point *ap,
 	return;
 }
 
-static int check_fstype_autofs_option(const char *options)
-{
-	char *tok, *tokbuf;
-	int found;
-
-	/*
-	 * Look for fstype= in options and return true if
-	 * the last occurrence is fstype=autofs.
-	 */
-	found = 0;
-	tokbuf = alloca(strlen(options) + 2);
-	strcpy(tokbuf, options);
-	tok = strtok_r(tokbuf, ",", &tokbuf);
-	if (tok) {
-		do {
-			if (strstr(tok, "fstype=")) {
-				if (strstr(tok, "autofs"))
-					found = 1;
-				else
-					found = 0;
-			}
-		} while ((tok = strtok_r(NULL, ",", &tokbuf)));
-	}
-
-	return found;
-}
-
 static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 			 const char *name, char *loc, char *options, void *ctxt)
 {
 	struct mapent *mm;
 	struct mapent *ro;
-	char t_dir[] = "/tmp/autoXXXXXX";
-	char *mnt_tmp_root, *mm_root, *mm_base, *mm_key;
-	const char *mnt_root, *target;
+	char *mm_root, *mm_base, *mm_key;
+	const char *mnt_root;
 	unsigned int mm_root_len, mnt_root_len;
 	int start, ret = 0, rv;
-	unsigned int move;
 
 	rv = 0;
 
-	if (!me || !me->multi) {
-		int loclen = strlen(loc);
-		int namelen = strlen(name);
-		const char *root = ap->path;
-
-		if (!strcmp(ap->path, "/-"))
-			root = name;
-
-		rv = sun_mount(ap, root, name, namelen, loc, loclen, options, ctxt);
-
-		goto done;
-	}
-
 	mm = me->multi;
 	mm_key = mm->key;
-	move = MOUNT_MOVE_NONE;
 
 	if (*mm_key == '/') {
 		mm_root = mm_key;
@@ -1157,22 +1087,12 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 	}
 	mm_root_len = strlen(mm_root);
 
-	mnt_tmp_root = NULL;
-	if (ap->flags & MOUNT_FLAG_REMOUNT) {
-		mnt_root = mm_root;
-		mnt_root_len = mm_root_len;
-	} else {
-		mnt_root = mkdtemp(t_dir);
-		if (!mnt_root)
-			return 1;
-		mnt_root_len = strlen(mnt_root);
-		mnt_tmp_root = (char *) mnt_root;
-	}
+	mnt_root = mm_root;
+	mnt_root_len = mm_root_len;
 
 	if (me == me->multi) {
 		/* name = NULL */
 		/* destination = mm_root */
-		target = mm_root;
 		mm_base = "/";
 
 		/* Mount root offset if it exists */
@@ -1189,17 +1109,13 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 				warn(ap->logopt,
 				      MODPREFIX "failed to parse root offset");
 				cache_delete_offset_list(me->mc, name);
-				goto error_out;
+				return 1;
 			}
-			ro_len = strlen(ro_loc);
+			ro_len = 0;
+			if (ro_loc)
+				ro_len = strlen(ro_loc);
 
-			if (!(ap->flags & MOUNT_FLAG_REMOUNT)) {
-				move = MOUNT_MOVE_OTHER;
-				if (check_fstype_autofs_option(myoptions))
-					move = MOUNT_MOVE_AUTOFS;
-			}
-
-			tmp = alloca(mnt_root_len + 1);
+			tmp = alloca(mnt_root_len + 2);
 			strcpy(tmp, mnt_root);
 			tmp[mnt_root_len] = '/';
 			tmp[mnt_root_len + 1] = '\0';
@@ -1208,7 +1124,8 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 			rv = sun_mount(ap, root, name, namelen, ro_loc, ro_len, myoptions, ctxt);
 
 			free(myoptions);
-			free(ro_loc);
+			if (ro_loc)
+				free(ro_loc);
 		}
 
 		if (ro && rv == 0) {
@@ -1217,32 +1134,25 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
 				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-				cleanup_multi_root(ap, mnt_root, mm_root, move);
-				goto error_out;
+				return 1;
 			}
 		} else if (rv <= 0) {
-			move = MOUNT_MOVE_NONE;
 			ret = mount_multi_triggers(ap, me, mm_root, start, mm_base);
 			if (ret == -1) {
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
 				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
-				goto error_out;
+				return 1;
 			}
 		}
 	} else {
 		int loclen = strlen(loc);
 		int namelen = strlen(name);
 
-		if (!(ap->flags & MOUNT_FLAG_REMOUNT)) {
-			move = MOUNT_MOVE_OTHER;
-			if (check_fstype_autofs_option(options))
-				move = MOUNT_MOVE_AUTOFS;
-		}
+		mnt_root = name;
 
 		/* name = mm_root + mm_base */
 		/* destination = mm_root + mm_base = name */
-		target = name;
 		mm_base = &me->key[start];
 
 		rv = sun_mount(ap, mnt_root, name, namelen, loc, loclen, options, ctxt);
@@ -1252,14 +1162,11 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
 				cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-				cleanup_multi_root(ap, mnt_root, mm_root, move);
-				goto error_out;
+				return 1;
 			}
 		} else if (rv < 0) {
 			char *mm_root_base = alloca(strlen(mm_root) + strlen(mm_base) + 1);
 	
-			move = MOUNT_MOVE_NONE;
-
 			strcpy(mm_root_base, mm_root);
 			strcat(mm_root_base, mm_base);
 
@@ -1268,25 +1175,15 @@ static int mount_subtree(struct autofs_point *ap, struct mapent *me,
 				error(ap->logopt, MODPREFIX
 					 "failed to mount offset triggers");
 				cleanup_multi_triggers(ap, me, mm_root, start, mm_base);
-				goto error_out;
+				return 1;
 			}
 		}
 	}
 
-	if (!move_mount(ap, mnt_root, target, move)) {
-		cleanup_multi_triggers(ap, me, mnt_root, start, mm_base);
-		cleanup_multi_root(ap, mnt_root, mm_root, move);
-		goto error_out;
-	}
-
-	if (mnt_tmp_root)
-		rmdir(mnt_tmp_root);
-
 	/* Mount for base of tree failed */
 	if (rv > 0)
 		return rv;
 
-done:
 	/*
 	 * Convert fail on nonstrict, non-empty multi-mount
 	 * to success
@@ -1295,12 +1192,6 @@ done:
 		rv = 0;
 
 	return rv;
-
-error_out:
-	if (mnt_tmp_root)
-		rmdir(mnt_tmp_root);
-
-	return 1;
 }
 
 /*
@@ -1322,11 +1213,11 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	char buf[MAX_ERR_BUF];
 	struct map_source *source;
 	struct mapent_cache *mc;
-	struct mapent *me = NULL;
+	struct mapent *me;
 	char *pmapent, *options;
 	const char *p;
 	int mapent_len, rv = 0;
-	int optlen, cur_state;
+	int cur_state;
 	int slashify = ctxt->slashify_colons;
 	unsigned int append_options;
 
@@ -1344,12 +1235,12 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 	macro_lock();
 
-	ctxt->subst = addstdenv(ctxt->subst);
+	ctxt->subst = addstdenv(ctxt->subst, NULL);
 
 	mapent_len = expandsunent(mapent, NULL, name, ctxt->subst, slashify);
 	if (mapent_len == 0) {
 		error(ap->logopt, MODPREFIX "failed to expand map entry");
-		ctxt->subst = removestdenv(ctxt->subst);
+		ctxt->subst = removestdenv(ctxt->subst, NULL);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
@@ -1359,7 +1250,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	if (!pmapent) {	
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 		logerr(MODPREFIX "alloca: %s", estr);
-		ctxt->subst = removestdenv(ctxt->subst);
+		ctxt->subst = removestdenv(ctxt->subst, NULL);
 		macro_unlock();
 		pthread_setcancelstate(cur_state, NULL);
 		return 1;
@@ -1367,7 +1258,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	pmapent[mapent_len] = '\0';
 
 	expandsunent(mapent, pmapent, name, ctxt->subst, slashify);
-	ctxt->subst = removestdenv(ctxt->subst);
+	ctxt->subst = removestdenv(ctxt->subst, NULL);
 
 	macro_unlock();
 	pthread_setcancelstate(cur_state, NULL);
@@ -1381,7 +1272,6 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		logerr(MODPREFIX "strdup: %s", estr);
 		return 1;
 	}
-	optlen = strlen(options);
 
 	p = skipspace(pmapent);
 
@@ -1446,7 +1336,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	if (check_is_multi(p)) {
 		char *m_root = NULL;
 		int m_root_len;
-		time_t age = time(NULL);
+		time_t age;
 		int l;
 
 		/* If name starts with "/" it's a direct mount */
@@ -1474,34 +1364,6 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			strcat(m_root, name);
 		}
 
-		cache_writelock(mc);
-		me = cache_lookup_distinct(mc, name);
-		if (!me) {
-			int ret;
-			/*
-			 * Not in the cache, perhaps it's a program map
-			 * or one that doesn't support enumeration
-			 */
-			ret = cache_add(mc, source, name, mapent, time(NULL));
-			if (ret == CHE_FAIL) {
-				cache_unlock(mc);
-				free(options);
-				return 1;
-			}
-		} else {
-			/*
-			 * If the entry exists it must not have any existing
-			 * multi-mount subordinate entries since we are
-			 * mounting this afresh. We need to do this to allow
-			 * us to fail on the check for duplicate offsets in
-			 * we don't know when submounts go away.
-			 */
-			cache_multi_writelock(me);
-			cache_delete_offset_list(mc, name);
-			cache_multi_unlock(me);
-		}
-		cache_unlock(mc);
-
 		cache_readlock(mc);
 		me = cache_lookup_distinct(mc, name);
 		if (me) {
@@ -1518,6 +1380,8 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			return 1;
 		}
 
+		age = me->age;
+
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 		cache_multi_writelock(me);
 		/* It's a multi-mount; deal with it */
@@ -1567,9 +1431,9 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			master_source_current_wait(ap->entry);
 			ap->entry->current = source;
 
-			status = add_offset_entry(ap, name,
-						m_root, m_root_len,
-						path, myoptions, loc, age);
+			status = update_offset_entry(ap, name,
+						     m_root, m_root_len,
+						     path, myoptions, loc, age);
 
 			if (status != CHE_OK) {
 				warn(ap->logopt, MODPREFIX "error adding multi-mount");
@@ -1579,20 +1443,25 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				free(path);
 				free(options);
 				free(myoptions);
-				free(loc);
+				if (loc)
+					free(loc);
 				pthread_setcancelstate(cur_state, NULL);
 				return 1;
 			}
 
-			free(loc);
+			if (loc)
+				free(loc);
 			free(path);
 			free(myoptions);
 		} while (*p == '/' || (*p == '"' && *(p + 1) == '/'));
 
 		/*
 		 * We've got the ordered list of multi-mount entries so go
-		 * through and set the parent entry of each
+		 * through and remove any stale entries if this is the top
+		 * of the multi-mount and set the parent entry of each.
 		 */
+		if (me == me->multi)
+			clean_stale_multi_triggers(ap, me, NULL, NULL);
 		cache_set_parents(me);
 
 		rv = mount_subtree(ap, me, name, NULL, options, ctxt);
@@ -1615,17 +1484,25 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		 * it's already been parsed (above) and any option string
 		 * has already been stripped so just use the remainder.
 		 */
+		cache_readlock(mc);
 		if (*name == '/' &&
 		   (me = cache_lookup_distinct(mc, name)) && me->multi) {
 			loc = strdup(p);
 			if (!loc) {
 				free(options);
+				cache_unlock(mc);
 				warn(ap->logopt, MODPREFIX "out of memory");
 				return 1;
 			}
-			loclen = strlen(p);
-			goto mount_it;
+			cache_multi_writelock(me);
+			rv = mount_subtree(ap, me, name, loc, options, ctxt);
+			cache_multi_unlock(me);
+			cache_unlock(mc);
+			free(loc);
+			free(options);
+			return rv;
 		}
+		cache_unlock(mc);
 
 		l = chunklen(p, check_colon(p));
 		loc = dequote(p, l, ap->logopt);
@@ -1638,13 +1515,13 @@ int parse_mount(struct autofs_point *ap, const char *name,
 		/* Location can't begin with a '/' */
 		if (*p == '/') {
 			free(options);
+			free(loc);
 			warn(ap->logopt,
 			      MODPREFIX "error location begins with \"/\"");
 			return 1;
 		}
 
-		if (!validate_location(loc)) {
-			warn(ap->logopt, MODPREFIX "invalid location %s", loc);
+		if (!validate_location(ap->logopt, loc)) {
 			free(loc);
 			free(options);
 			return 1;
@@ -1669,9 +1546,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 				return 1;
 			}
 
-			if (!validate_location(ent)) {
-				warn(ap->logopt,
-				     MODPREFIX "invalid location %s", loc);
+			if (!validate_location(ap->logopt, ent)) {
 				free(ent);
 				free(loc);
 				free(options);
@@ -1700,29 +1575,38 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			p = skipspace(p);
 		}
 
-		loclen = strlen(loc);
-		if (loclen == 0) {
-			free(loc);
-			free(options);
-			error(ap->logopt,
-			      MODPREFIX "entry %s is empty!", name);
-			return 1;
+		/*
+		 * If options are asking for a hosts map loc should be
+		 * NULL but we see it can contain junk, so ....
+		 */
+		if ((strstr(options, "fstype=autofs") &&
+		     strstr(options, "hosts"))) {
+			loc = NULL;
+			loclen = 0;
+		} else {
+			loclen = strlen(loc);
+			if (loclen == 0) {
+				free(loc);
+				free(options);
+				error(ap->logopt,
+				      MODPREFIX "entry %s is empty!", name);
+				return 1;
+			}
 		}
-mount_it:
+
 		debug(ap->logopt,
 		      MODPREFIX "core of entry: options=%s, loc=%.*s",
 		      options, loclen, loc);
 
-		cache_readlock(mc);
-		cache_multi_writelock(me);
-
-		rv = mount_subtree(ap, me, name, loc, options, ctxt);
+		if (!strcmp(ap->path, "/-"))
+			rv = sun_mount(ap, name, name, name_len,
+				       loc, loclen, options, ctxt);
+		else
+			rv = sun_mount(ap, ap->path, name, name_len,
+				       loc, loclen, options, ctxt);
 
 		free(loc);
 		free(options);
-
-		cache_multi_unlock(me);
-		cache_unlock(mc);
 		pthread_setcancelstate(cur_state, NULL);
 	}
 	return rv;
@@ -1733,10 +1617,12 @@ int parse_done(void *context)
 	int rv = 0;
 	struct parse_context *ctxt = (struct parse_context *) context;
 
+	instance_mutex_lock();
 	if (--init_ctr == 0) {
 		rv = close_mount(mount_nfs);
 		mount_nfs = NULL;
 	}
+	instance_mutex_unlock();
 	if (ctxt)
 		kill_context(ctxt);
 
